local set = krework.chars.tails.settings

local function boostHandler(player)
	local mo = player.mo
	
	-- Calculate boost to apply
	local ascendspeed = set.BOOST_UPSPEED
	if mo.eflags&MFE_UNDERWATER
		ascendspeed = $ * 2/3
	end
	
	P_SetObjectMomZ(mo, ascendspeed, true)
	
	-- Dust clouds (or bubbles) to represent Tails boosting
	local source
	local dusttype
	local dust1x
	local dust1y
	local dust2x
	local dust2y
	local dustz
	local dustscale
	local dustcolor
	
	-- Spawn bubbles behind Tails's feet when swimming
	if mo.eflags&MFE_UNDERWATER
		source = player.mo
		
		-- Random deviation from center
		local bubblewobble = 2 * FRACUNIT / 5
		-- Need this because P_RandomRange can't handle a range >=65536
		local bubblewobblemult = 2
		
		local angle = (source.angle + ANGLE_157h)
		dust1x = FixedMul(cos(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
		dust1y = FixedMul(sin(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
		dust2x = FixedMul(cos(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
		dust2y = FixedMul(sin(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
		dustz = 0
		
		dusttype = MT_MEDIUMBUBBLE
		
		dustscale = false
		dustcolor = false
	-- Spawn dust clouds above Tails's tails when flying
	else
		source = player.followmobj
		if not source
			return
		end
		
		-- Tails's tails have a 4-frame animation that rotates 180 degrees
		local angle = (player.mo.angle + (source.frame * ANGLE_45))
		dust1x = FixedMul(cos(angle), source.radius)
		dust1y = FixedMul(sin(angle), source.radius)
		dust2x = FixedMul(cos(angle+ANGLE_180), source.radius)
		dust2y = FixedMul(sin(angle+ANGLE_180), source.radius)
		dustz = source.height*4/5
		
		dusttype = MT_SPINDUST
		
		dustscale = set.BOOST_SCALE_DUST
		dustcolor = set.BOOST_COLOR_DUST
	end
	
	local d1 = P_SpawnMobjFromMobj(source, dust1x, dust1y, dustz, dusttype)
	local d2 = P_SpawnMobjFromMobj(source, dust2x, dust2y, dustz, dusttype)
	
	-- Change color to represent spending boosts
	if dustcolor and mo.krework.tails.flycount < 3
		local dcolor = set.BOOST_DUST_COLORS[mo.krework.tails.flycount] or set.BOOST_DUST_DEFAULT
		d1.color, d2.color = dcolor, dcolor
		d1.colorized, d2.colorized = true, true
	end
	
	-- Reduce in size to represent spending boosts
	if dustscale
		d1.scale = $ * (mo.krework.tails.flycount+1) / 3
		d2.scale = d1.scale
	end
end

local function hoverHandler(player)
	local mo = player.mo
	
	-- Declare flight-related variables
	local time = mo.krework.tails.flytime
	local hoverspeed = set.FLY_HOVERSPEED
	local flip = P_MobjFlip(mo)
	local speed = FixedHypot(mo.momx, mo.momy)
	local ascendtics = set.FLY_BOOST_TIME
	
	-- Normal flying stuff
	if mo.state == S_PLAY_FLYHOVER
		if time % 10 == 1
			S_StartSound(mo, sfx_s3kba)
		end
		
	-- Tired flying differences
	elseif mo.state == S_PLAY_FLYHOVER_TIRED
		hoverspeed = set.FLY_TIRED_HOVERSPEED
		
		if time % 5 == 1
			S_StartSound(mo, sfx_s3kbb)
		end
		
	-- Swimming changes
	elseif mo.state == S_PLAY_SWIMHOVER
		ascendtics = set.SWIM_BOOST_TIME
		
	-- What
	else
		mo.state = S_PLAY_FALL
		return
	end
	
	-- Fall no faster than your hover speed
	if (flip == 1 and mo.momz < hoverspeed) or (flip == -1 and mo.momz > -hoverspeed)
		P_SetObjectMomZ(mo, hoverspeed, false)
	end
	
	-- If you're still in vertboost, apply boost
	if (time <= ascendtics)
		boostHandler(player)
	end
	
	mo.krework.tails.flytime = $ + 1
end

rawset(_G.krework.chars.tails, "preFrame", function(player)
	-- Get outta here, Tailsbot
	if player.bot
		return
	end
	
	local mo = player.mo
	
	if player.charability == CA_FLY
		player.charability = CA_NONE
	end
	
	-- Infinite vertboost count when swimming
	if (mo.eflags & MFE_UNDERWATER)
		mo.krework.tails.flycount = set.BOOST_COUNT
	end
	
	-- If you have boosts left, you can boost
	player.krework.jumpused = not(max(mo.krework.tails.flycount, 0))
	
	-- Unless you're in boost cooldown
	if (mo.state == S_PLAY_FLYHOVER
	or mo.state == S_PLAY_FLYHOVER_TIRED
	or mo.state == S_PLAY_SWIMHOVER)
		-- Move tails faster to show that you're in boost cooldown
		if mo.krework.tails.flytime < set.FLY_COOLDOWN
			player.followmobj.tics = max($-1, 1)
			player.krework.jumpused = true
		end
	end
	
	if player.mo.state == S_PLAY_TAILSCATCH
		local thrown = mo.krework.tails.throwplayer
		
		mo.krework.tails.throwup = (thrown.cmd.buttons & BT_JUMP)
		mo.krework.tails.throwdown = (thrown.cmd.buttons & BT_USE)
		thrown.cmd.buttons = 0
		thrown.cmd.sidemove = 0
		thrown.cmd.forwardmove = 0
		
		player.cmd.buttons = 0
	end
end)

addHook("AbilitySpecial", function(player)
	if not krework.checkExistence(player)
	or player.mo.skin != "tails"
	or player.bot
	or not (player.pflags & PF_JUMPED)
		return false
	end
		
	if player.krework.jumpused
		S_StartSound(player.mo, sfx_pudpud)
		return true
	end
	
	if (player.mo.krework.tails.flytime > 0 and player.mo.krework.tails.flytime < set.FLY_COOLDOWN)
		return true
	end
	
	player.mo.state = S_PLAY_FLYBOOST
	return true
end)

rawset(_G.krework.chars.tails, "playerThink", function(player)
	if player.bot
		return
	end
	
	local mo = player.mo
	
	if mo.state == S_PLAY_TAILSCATCH
		local thrown = mo.krework.tails.throwplayer
		
		mo.momx = 0
		mo.momy = 0
		mo.momz = 0
		
		P_TeleportMove(thrown.mo, mo.x, mo.y, mo.z)
		thrown.mo.momx = 0
		thrown.mo.momy = 0
		thrown.mo.momz = 0
		thrown.mo.state = S_PLAY_TAILSCAUGHT
		A_KRResetAbilities(thrown)
		
		mo.krework.tails.throwtime = $ + 1
	end
	
	-- Hover -> Swim
	if (mo.state == S_PLAY_FLYHOVER or mo.state == S_PLAY_FLYHOVER_TIRED) and (mo.eflags & MFE_UNDERWATER)
		mo.state = S_PLAY_SWIMHOVER
	end
	
	-- Swimming -> hover
	if (mo.state == S_PLAY_SWIMHOVER) and not (mo.eflags & MFE_UNDERWATER)
		mo.state = S_PLAY_FLYHOVER
		mo.krework.tails.flytime = min($, set.FLY_BOOST_TIME + 1)
	end
	
	-- Hover -> tired
	if mo.state == S_PLAY_FLYHOVER and mo.krework.tails.flytime > set.FLY_TIRED_TIME
		mo.state = S_PLAY_FLYHOVER_TIRED
	end
	
	-- Tired -> fall
	if mo.state == S_PLAY_FLYHOVER_TIRED and mo.krework.tails.flytime > set.FLY_END_TIME
		mo.state = S_PLAY_FALL
	end
	
	-- If you're flying, run flying physics
	if mo.state == S_PLAY_FLYHOVER
	or mo.state == S_PLAY_FLYHOVER_TIRED
	or mo.state == S_PLAY_SWIMHOVER
		hoverHandler(player)
		
	-- Otherwise, reset how long you've been flying
	else
		mo.krework.tails.flytime = 0
	end
	
	if not P_IsObjectOnGround(mo)
	and not player.krework.jumpused
	and mo.state != S_PLAY_TAILSCATCH
	and mo.krework.tails.throwtime == 0
		local nearest = krework.getNearbyAlly(mo, set.MAX_GRAB_RANGE)
		if nearest
			local fistbump = P_SpawnMobj(0, 0, 0, MT_FISTBUMP)
			if splitscreen
				P_TeleportMove(fistbump, (player.mo.x + nearest.mo.x)/2, (player.mo.y + nearest.mo.y)/2, (player.mo.z + nearest.mo.z)/2 + (48 * FRACUNIT))
			else
				if player == consoleplayer
					P_TeleportMove(fistbump, nearest.mo.x, nearest.mo.y, nearest.mo.z + nearest.mo.height + (16 * FRACUNIT))
				elseif nearest == consoleplayer
					P_TeleportMove(fistbump, player.mo.x, player.mo.y, player.mo.z + player.mo.height + (16 * FRACUNIT))
				else
					fistbump.flags2 = $ | MF2_DONTDRAW
				end
			end
		end
	end
end)
